import {EditorState, Plugin} from 'prosemirror-state'
import {EditorView} from 'prosemirror-view'
import {history} from 'prosemirror-history'
import {keymap} from 'prosemirror-keymap'
import {baseKeymap} from 'prosemirror-commands'
import {DOMParser, DOMSerializer} from 'prosemirror-model'

import {Menu, blocks, marks} from './godo-menus.mjs'
import {dialog} from './godo-dialog.mjs'
import {loadSchema, buildKeymap} from './utils.mjs'
import base from './components/base.mjs'
import fontMarks from './components/font-marks.mjs'
import heading from './components/heading.mjs'
import link from './components/link.mjs'
import list from './components/list.mjs'

function menuPlugin (menu) {
  return new Plugin({
    view (editorView) {
      let menuView = new Menu(menu, editorView)
      editorView.dom.parentNode.insertBefore(menuView.el, editorView.dom)
      return menuView
    },
  })
}

function changeEventPlugin (editor) {
  return new Plugin({
    state: {
      init () {},
      apply (tr, callback) {
        if (tr.docChanged) {
          clearTimeout(callback)
          callback = setTimeout(() => {
            editor.textarea.value = editor.getHTML(tr.doc.content)
            editor.textarea.dispatchEvent(new Event('change', { 'bubbles': true }))
          }, 300)
        }
        return callback
      },
    },
  })
}

export default class Godo extends window.HTMLElement {
  static observedAttributes = ['linked-input']

  constructor () {
    super()
    this.linkedInput = null

    // must keep a reference on those handlers to be able to remove them
    this.onUpdateEvent = () => {
      this.view.updateState(this._createState(this.linkedInput.value))
    }

    this.onLinkedInputFormSubmit = () => {
      this.linkedInput.value = this.getHTML(this.view.state.doc.content)
    }
  }

  attributeChangedCallback (name, oldValue, newValue) {
    if (name === 'linked-input') {
      this.updateLinkedInput(newValue)
    }
  }

  updateLinkedInput (inputId) {
    if (!this.isConnected) {
      return
    }

    const updateEvent = this.getAttribute('update-event') || false

    const oldInput = this.linkedInput
    const newInput = inputId && document.querySelector(`#${inputId}`)

    if (oldInput === newInput) {
      return
    }

    if (oldInput) {
      if (oldInput.form) {
        oldInput.form.removeEventListener('submit', this.onLinkedInputFormSubmit)
      }

      if (updateEvent) {
        oldInput.removeEventListener(updateEvent, this.onUpdateEvent)
      }
    }

    if (newInput) {
      if (updateEvent) {
        newInput.addEventListener(updateEvent, this.onUpdateEvent)
      }

      if (newInput.form) {
        newInput.form.addEventListener('submit', this.onLinkedInputFormSubmit)
      }

      if (this.view) {
        newInput.value = this.getHTML(this.view.state.doc.content)
      }
    }

    this.linkedInput = newInput
  }


  connectedCallback () {
    const initialContent = this.innerHTML
    this.innerHTML = ''

    const options = {
      schema: this.getAttribute('schema') || 'full',
      instantUpdate: this.getAttribute('instant-update') !== null,
      updateEvent: this.getAttribute('update-event') || false,
    }

    if (this.getAttribute('heading-levels')) {
      options.headingLevels = this.getAttribute('heading-levels')
    }

    const componentsDefinitions = [base, fontMarks, link]
    if (options.schema === 'full') {
      componentsDefinitions.push(heading, list)
    }

    const components = componentsDefinitions.map(c => c(options))
    this.schema = loadSchema(components)

    this.editorWrapper = document.createElement('div')
    this.editorWrapper.className = 'godo'
    this.appendChild(this.editorWrapper)

    this.marksMenu = menuPlugin(marks(components, this.schema))
    this.blocksMenu = menuPlugin(blocks(components, this.schema))
    this.menuDialog = document.body.appendChild(dialog)

    this.pluginsList = [
      history(),
      keymap(buildKeymap(components, this.schema, null)),
      keymap(baseKeymap),
      this.marksMenu,
    ]

    if (options.schema !== 'basic' && !this.readOnly) this.pluginsList.push(this.blocksMenu)
    if (options.instantUpdate) this.pluginsList.push(changeEventPlugin(this))

    const label = this.id ? document.querySelector(`label[for=${this.id}]`) : undefined
    this.view = new EditorView(this.editorWrapper, {
      state: this._createState(initialContent),
      editable: () => !this.readOnly,
      attributes: {
        class: 'godo--editor',
        role: 'textbox',
        'aria-multiline': 'true',
        'aria-labelledby': label ? label.id : undefined,
      },
    })

    this.updateLinkedInput(this.getAttribute('linked-input'))
  }

  _createState (content) {
    const w = document.createElement('div')
    w.innerHTML = content
    return EditorState.create({
      schema: this.schema,
      doc: DOMParser.fromSchema(this.schema).parse(w),
      plugins: this.pluginsList,
    })
  }

  getHTML (from) {
    const div = document.createElement('div')
    const fragment = DOMSerializer
      .fromSchema(this.schema)
      .serializeFragment(from)
    div.appendChild(fragment)
    return div.innerHTML
  }
}

window.customElements.define('godo-editor', Godo)
